
/* cnode.h -- definitions for kernel-internal cnode struct,
   used to keep track of inodes on the Coda filesystem

  Copyright 1997-98 Michael Callahan
  Copyright 1999 Carnegie Mellon University
  This program is free software.  You may copy it according
  to the conditions of the GNU General Public License version 2.0
  or later; see the file COPYING in the source directory.
*/

#ifndef	_CNODE_H_
#define	_CNODE_H_

#define CODA_CNODE_MAGIC        0x47114711

/*
 * Cnode lookup stuff.
 * NOTE: CNODE_CACHESIZE must be a power of 2 for cfshash to work!
 */
#define CNODE_CACHESIZE 512
#define cnode_hash(fid) \
    (((fid)->Volume + (fid)->Vnode) & (CNODE_CACHESIZE-1))


/* defintion of cnode, which combines ViceFid with inode information */

struct cnode {
        u_short	    c_flags;	     /* flags (see below) */
        ViceFid	    c_fid;	     /* file handle */
        int             c_magic;     /* to verify the data structure */
        u_short	        c_ocount;    /* count of openers */
        u_short         c_owrite;    /* count of open for write */
        struct coda_vattr    c_vattr;     /* attributes */
        char            *c_symlink;  /* pointer to symbolic link */
        u_short         c_symlen;    /* length of symbolic link */
        struct cnode    *c_next;     /* next cnode in the cache */
        dev_t	        c_device;    /* associated vnode device */
        ino_t	        c_inode;     /* associated vnode inode "number?"?? */
};

/* flags */
#define C_VATTR       0x1         /* Validity of vattr in the cnode */
#define VALID_VATTR(cp)          ((cp->c_flags) & C_VATTR)

#define C_SYMLINK     0x2         /* Validity of symlink pointer in the cnode */
#define VALID_SYMLINK(cp)        ((cp->c_flags) & C_SYMLINK)


#define C_DYING	      0x4	  /* Set for outstanding cnodes from last venus (which died) */
#define IS_DYING(cp)		 ((cp->c_flags) & C_DYING)

#define CN_WANTED     0x8        /* Set if lock wanted */
#define CN_LOCKED     0x10       /* Set if lock held */
#define CN_UNMOUNTING 0X20       /* Set if unmounting */
#define IS_UNMOUNTING(cp)       ((cp)->c_flags & CN_UNMOUNTING)

#endif	not _CNODE_H_

