/* readdir.c -- routines for reading Coda directory files
   Copyright 1997-98 Michael Callahan
   Copyright 1999 Carnegie Mellon University
   This program is free software.  You may copy it according
   to the conditions of the GNU General Public License version 2.0
   or later; see the file COPYING in the source directory.
*/

#include "MC.h"

#pragma VxD_LOCKED_DATA_SEG
#pragma VxD_LOCKED_CODE_SEG

#include <string.h>
#pragma intrinsic (memcpy, strcat, strlen, strcpy)

int do_potemkin_readdir(PREADDIR p){
	int res = 0;

	if (p->hdl) {// 0 means not first time, -1 means first time
		res = venus_open_by_path(p->VFid, (char *)&p->buf, 0, C_O_READ);
		if (!res)
			p->hdl=0; // indicates, that venus has opened the file by path
			logmsg("got path");
			strcpy((char*)&p->buf[strlen(p->buf)-1], "*.*");
			logmsg(p->buf);
			res = R0_FindFirstFile(0, p->buf, (_WIN32_FIND_DATA*)&(p->bufstart), &p->pos); 
			//pos has find handle for FindNextFile and bufstart contains adress of Find Data
			if (res) printf("error in R0_FindFirstFile %x", res);
	}
	else{
		res = R0_FindNextFile(p->pos, (_WIN32_FIND_DATA*)&(p->bufstart));
		if (res == ERROR_NO_MORE_FILES){
			return res;
		}
	}

	if (res){
		logmsg("File Corrupt");
		return ERROR_FILE_CORRUPT;
	}

	return 0;
}
 
int opendir (ViceFid VFid, PREADDIR *p)
{
	int res;

	logmsg("--opendir--");

	if (!*p) {
		*p = _HeapAllocate (sizeof(READDIR), 0);

		if (!*p)
			return ERROR_NOT_ENOUGH_MEMORY;

		(*p)->VFid= VFid;
		(*p)->hdl = 0;
		(*p)->pos = 0;
		(*p)->bufstart = 0;
		(*p)->bufend = 0;
	}

	printf ("VFID: %x %x %x", VFid.Volume,VFid.Vnode, VFid.Unique);
	res = openVFid (VFid, C_O_READ, 
			(WORD) ACCESS_READONLY, (WORD) 0, (BYTE) ACTION_OPENEXISTING, 
			NULL, &(*p)->hdl, NULL, NULL);
	if (res) {
		printf ("open dir failed : %x", res);
		_HeapFree (*p, 0);
		*p = NULL;
		return res;
	}

	res = R0_GetFileSize((*p)->hdl, &(*p)->filesize);
	if (res) {
		printf ("get file size failed: %x", res);
		closedir(*p);
		_HeapFree (*p, 0);
		*p = NULL;
	}

	logmsg("--end of opendir--");

	return res;
}

static int getbytes (PREADDIR preaddir, DWORD n, char *p) 
{
	DWORD len, len2, res, readlen;
	printf ("getbytes[%x]: pos %x buf [%x,%x) len %x", (DWORD) preaddir,
		preaddir->pos, preaddir->bufstart, preaddir->bufend, n);
	len = n;
	if (len > preaddir->bufend - preaddir->pos)
		len = preaddir->bufend - preaddir->pos;
	memcpy (p, preaddir->buf + preaddir->pos - preaddir->bufstart, len);
	p += len;
	preaddir->pos += len;
	printf ("getbytes: p %x, pos %x ", p, preaddir->pos);

	if (preaddir->pos < preaddir->bufend){
		printf ("getbytes: len %x ", len);
		return len;
	}

	readlen = READDIR_BUFSIZE;
	if (readlen > preaddir->filesize - preaddir->pos)
		readlen = preaddir->filesize - preaddir->pos;

	if (readlen)
		res = R0_ReadFile(TRUE,
			preaddir->hdl,
			readlen,
			preaddir->pos,
			preaddir->buf,
			(PDWORD) &readlen);
	else
		res = 0;
	preaddir->bufstart = preaddir->pos;
	preaddir->bufend = preaddir->bufstart + readlen;
	printf ("getbytes: res %x readlen %x", res, readlen);
	if (res) {
		printf ("getbytes: res %x ", res);
		return len;
	}
	len2 = n - len;
	if (len2 > readlen)
		len2 = readlen;
	memcpy (p, preaddir->buf, len2);
	preaddir->pos += len2;
	printf ("getbytes: return %x ", len+len2);
	return len + len2;
}

static void skip(PREADDIR preaddir, int n)
{
	preaddir->pos += n;
}

#define GETBYTES(preaddir, n, p) ((preaddir->pos + n < preaddir->bufend) ? \
	memcpy(p, preaddir->buf + (preaddir->pos - preaddir->bufstart), n), \
		preaddir->pos += n, n : \
	getbytes(preaddir, n, p))

int readdir (PREADDIR p, PCHAR q, PDWORD plen)
{
	int len;
	struct venus_dirent deh;
	int deHdrLen = (char *) (&deh.d_name) - (char *) (&deh);

start:

	len = GETBYTES(p, deHdrLen, (char *) &deh);
	printf ("readdir: pos %x bufstart %x bufend %x", p->pos, p->bufstart, p->bufend);
	logmsg ("readdir: out first GETBYTES");

	if (deh.d_namlen == 0 && deh.d_reclen+p->pos-len > p->bufend ) {
		/* end of directory */
		printf ("readdir: reclen %x namlen %x fileno %x", deh.d_reclen, deh.d_namlen, deh.d_fileno);
		logmsg ("readdir: > bufend");
		*plen = 0;
		return 0;
	}
	
	if (len	!= deHdrLen) {
		logmsg ("error getting dirent header");
		return ERROR_FILE_CORRUPT;
	}

	printf ("getbytes: d_namelen %x ", deh.d_namlen);
	if (GETBYTES(p, deh.d_namlen, q) != deh.d_namlen) {
		logmsg ("error getting dirent name");
		return ERROR_FILE_CORRUPT;
	}
	printf ("readdir: pos %x bufstart %x bufend %x", p->pos, p->bufstart, p->bufend);

	*plen = deh.d_namlen;
	q[deh.d_namlen] = '\0';

	skip (p, deh.d_reclen - deh.d_namlen - deHdrLen);
	return 0;
}

void closedir(PREADDIR p)
{
	int res = 0;

	logmsg("--closedir--");
	printf("closedir: close handle: %x", (p) ? p->hdl : NULL);
	if (p){
		logmsg("--about R0_CloseFile--");
//		lock = 1;
		if (p->hdl) res = R0_CloseFile(p->hdl);
//		lock = 0;
		logmsg("--after R0_CloseFile--");
		if (res) {
			printf ("R0_CloseFile Error in closedir: %x", res);
			goto error;
		}
		if (p->hdl) res = venus_close(p->VFid, C_O_READ); 
	}

error:

	printf("--end of closedir res: %x--", res);
}