/*

  init.c -- initialization code for Coda Minicache
  Copyright 1997-98 Michael Callahan
  Copyright 1999 Carnegie Mellon University
  This program is free software.  You may copy it according
  to the conditions of the GNU General Public License version 2.0
  or later; see the file COPYING in the source directory.

  Thrown away after startup.

*/


#define WANTVXDWRAPS
#include <basedef.h>
#include <vmm.h>
#include <ios.h>
#include <ifs.h>
#include <vwin32.h>
#include <vxdwraps.h>

#include <winerror.h>

#include "ifsmgr.h"
#include "iosdcls.h"


#include <string.h>
#pragma intrinsic (memset)

#pragma VxD_LOCKED_DATA_SEG
#pragma VxD_LOCKED_CODE_SEG

static char copyright[] = "Copyright 1997 Michael Callahan\n";


UINT uiProviderId = 0x001b0000;
UINT uiNetId = 0;

int (*RealConnectNetResource) (pioreq pir) = NULL;

int ConnectNetResource (pioreq pir)
{
	if (RealConnectNetResource)
		return pir->ir_error = (*RealConnectNetResource) (pir);
	return pir->ir_error = ERROR_UNEXP_NET_ERR;
}

DWORD MCSTUB_Get_ConnectNetResource ()
{
	return (DWORD) RealConnectNetResource;
}

DWORD MCSTUB_Set_ConnectNetResource (int (*f) (pioreq))
{
	int (*old) (pioreq);

	old = RealConnectNetResource;
	RealConnectNetResource = f;
	return (DWORD) old;
}

void (*RealAsyncEvent) (PAEP);

void OnAsyncEvent (PAEP pAEP)
{
	if (RealAsyncEvent)
		(*RealAsyncEvent) (pAEP);
	else
		pAEP->AEP_result = AEP_SUCCESS;
}

DWORD MCSTUB_Get_AsyncEvent ()
{
	return (DWORD) RealAsyncEvent;
}

DWORD MCSTUB_Set_AsyncEvent (void (*f) (PAEP))
{
	void (*old) (PAEP);

	old = RealAsyncEvent;
	RealAsyncEvent = f;
	return (DWORD) old;
}

DWORD MCSTUB_Get_ProviderId ()
{
	return uiProviderId;
}

#pragma VxD_INIT_DATA_SEG
#pragma VxD_INIT_CODE_SEG

/* symbols defined in MCvxd.asm */
extern DRP MC_DRP;
extern DWORD BSSstart, BSSend;

void OnAsyncEvent(PAEP);

BOOL OnDeviceInit ()
{
	/* clear bss segment */
	memset (&BSSstart, 0, ((char *)&BSSend) - ((char *)&BSSstart));

	/* check that IOS is loaded */
	if (!IOS_Get_Version())
		return FALSE;

	/* register with IOS

	   We have to supply an async event routine.  If we 
	   registered as a FSD (filesystem driver), then this 
	   would be unnecessary.  However, according to Oney, 
	   Microsoft's FSDs register as TSD's (type-specific
	   drivers).  These _do_ receive async events from 
	   the IOS.

	   Go figure.
    */
    MC_DRP.DRP_aer = (PVOID) OnAsyncEvent;
	MC_DRP.DRP_feature_code |= DRP_FC_4_SECS;
	IOS_Register(&MC_DRP);
	if ((MC_DRP.DRP_reg_result == DRP_ABORT) ||
		(MC_DRP.DRP_reg_result == DRP_INVALID_LAYER))
		return FALSE;

	/* register with IFSMgr */
	uiProviderId = IFSMgr_RegisterNet (ConnectNetResource, 
		IFSMGRVERSION, uiNetId);

	if (uiProviderId == -1)
		return FALSE;

	return TRUE;
}

