/*  
  Copyright 1997-98 Michael Callahan
  Copyright 1999 Carnegie Mellon University
  This program is free software.  You may copy it according
  to the conditions of the GNU General Public License version 2.0
  or later; see the file COPYING in the source directory.
*/

#include "sock.h"

typedef struct _UDPPACKET {
	DWORD	udp_ready;
	DWORD	udp_size;
	DWORD	udp_fromip;
	DWORD	udp_fromport;
} UDPPACKET, *PUDPPACKET;

int
sys_sendto (HVM hVM, PCRS pRegs, PSOCKET pSock)
{
	TDI_REQUEST r;
	PTRANSPORT_ADDRESS pTA = MakeTransportAddress(pRegs->Client_EBX, (WORD) pRegs->Client_EDX);
	PCALLBACKMEMO pCBM = NewCallback();
	PTDI_CONNECTION_INFORMATION pTCI = _HeapAllocate(
				sizeof(TDI_CONNECTION_INFORMATION), HEAPZEROINIT);
	PNDIS_BUFFER pNB = _HeapAllocate(
				sizeof(NDIS_BUFFER), HEAPZEROINIT);
	DWORD pLinBuf, nPages, pLockedBuf, nBytesSent;
	int res = 0;
	TDI_STATUS err;

	if (!(pSock->s_flags & SOCKFL_BOUND)) {
		res = -SOCKERR_NOT_BOUND;
		goto error;
	}

	if (!pTA || !pCBM || !pTCI || !pNB) {
		res = -SOCKERR_NO_MEMORY;
		goto error;
	}

	pLinBuf = Client_Ptr_Flat(CLIENT_DS, CLIENT_SI);
	nPages = ((pLinBuf + pRegs->Client_ECX + 0xfff) >> 12) - (pLinBuf >> 12) + 1;
	pLockedBuf = _LinPageLock (pLinBuf >> 12, nPages, PAGEMAPGLOBAL);
	if (!pLockedBuf) {
		res = -SOCKERR_ACCESS;
		goto error;
	}

	pTCI->RemoteAddressLength = IP_TRANSPORT_ADDRESS_LENGTH + 4;
	pTCI->RemoteAddress = (PVOID) pTA;

	memcpy((PVOID) &pNB->Signature, "NBUF", 4);
	pNB->VirtualAddress = (PVOID) (pLockedBuf + (pLinBuf & 0xfff));
	pNB->Length = pRegs->Client_ECX;

	r.Handle.AddressHandle = pSock->s_addrobj->addr_hdl;
	r.RequestNotifyObject = &SignalCallback;
	r.RequestContext = (PVOID) pCBM;
	err = pDT->TdiSendDatagramEntry (&r, pTCI, pNB->Length, &nBytesSent, pNB);
	if (err == TDI_PENDING)
		err = WaitCallback(pCBM);
	res = -err;
	_LinPageUnlock(((DWORD) pLockedBuf) >> 12, nPages, PAGEMAPGLOBAL);

error:
	if (pCBM) FreeCallback(pCBM);
	if (pTA) FreeTransportAddress(pTA);
	if (pTCI) _HeapFree(pTCI, 0);
	if (pNB) _HeapFree(pNB, 0);
	return res;
}


int
sys_recvfrom(HVM hVM, PCRS pRegs, PSOCKET pSock)
{
	PCHAR pLinBuf = (PCHAR) Client_Ptr_Flat(CLIENT_DS, CLIENT_SI);
	UDPPACKET udp;
	DWORD len;
	int gotlen;
	int res = 0;

	if (!pLinBuf)
		return -SOCKERR_ACCESS;

	_asm pushf
	_asm cli
	while (!pSock->s_recvpackets) {
		_asm popf
		/* no packet.  either return WOULD_BLOCK or wait */
		if (!(pSock->s_flags & SOCKFL_NONBLOCK)) {
			return -SOCKERR_WOULD_BLOCK;
		}
		wait (pSock, POLL_READ);
		_asm pushf
		_asm cli
	}
	pSock->s_recvpackets--;
	_asm popf

	LockCBuf (pSock->s_recvbuf);

	GetCBuf (pSock->s_recvbuf, (PCHAR) &udp, sizeof(udp));
	if (udp.udp_ready != 0x80)
		res = - (int) udp.udp_ready;
	pRegs->Client_EDX = udp.udp_fromport; 
	pRegs->Client_EBX = udp.udp_fromip;

	len = udp.udp_size;
	if (len > pRegs->Client_ECX)
		len = pRegs->Client_ECX;
	gotlen = GetCBuf (pSock->s_recvbuf, pLinBuf, len);
	if (gotlen != (int) len) {
		UnlockCBuf (pSock->s_recvbuf);
		pRegs->Client_ECX = (gotlen << 16) | len;
		return -SOCKERR_INTERNAL;
	}
	if (len < udp.udp_size)
		PopCBuf(pSock->s_recvbuf, udp.udp_size - len);
	pRegs->Client_ECX = len;

	UnlockCBuf (pSock->s_recvbuf);

	if (res)
		return res;
	else
		return len;
}

struct RcvCallback {
	PSOCKET				cb_socket;
	PCHAR				cb_ready;
	EventRcvBuffer		cb_event;
	NDIS_BUFFER			cb_NB1;
	NDIS_BUFFER			cb_NB2;
	WORD				cb_flags;
};

void
RcvCallback (PVOID context, TDI_STATUS final, uint ByteCount) 
{
	struct RcvCallback *pCBM = (struct RcvCallback *)context;

	if (!pCBM)
		return;

	if (final != TDI_SUCCESS) {
		/* XXX ack!  what to do??? this should never happen */
		*pCBM->cb_ready = (char) final;
	}

	*pCBM->cb_ready = (char) 0x80;

	_asm pushf
	_asm cli
	pCBM->cb_socket->s_recvpackets++;
	_asm popf

	if (pCBM->cb_socket->s_post) {
		(*pCBM->cb_socket->s_post)++;
		_SignalID((DWORD) pCBM->cb_socket->s_post);
	}
	_HeapFree (pCBM, 0);
}


TDI_STATUS 
RcvDatagram(PVOID EventContext, 
			uint cbAddr, PTRANSPORT_ADDRESS pAddr, 
			uint cbOpt, PVOID pOpt, 
			uint dwFlags,
			uint cbIndic, uint cbAvail, 
			uint *pcbTaken, uchar *pData, 
			EventRcvBuffer **ppBuffer)
{
	PSOCKET pSock = (PSOCKET) EventContext;
	UDPPACKET udp;
	PTDI_ADDRESS_IP pIpAddr = (PTDI_ADDRESS_IP)
			(((PCHAR) pAddr) + 4 + 2 + 2);
	struct RcvCallback *pCBM;

	if (!pSock || pSock->s_signature != SOCKSIG)
		return TDI_NOT_ACCEPTED;

	if (!TryLockCBuf (pSock->s_recvbuf)) {
		return TDI_NOT_ACCEPTED;
	}

	/* ok, have cbuf locked */
	if ((uint) AvailCBuf (pSock->s_recvbuf) <= sizeof(udp) + cbAvail) {
		/* no space */
		UnlockCBuf (pSock->s_recvbuf);
		return TDI_NOT_ACCEPTED;
	}

	/* make NDIS_BUFFER header(s), etc */
	pCBM = _HeapAllocate(sizeof(struct RcvCallback), HEAPZEROINIT);
	if (!pCBM) {
		UnlockCBuf (pSock->s_recvbuf);
		return TDI_NOT_ACCEPTED;
	}

	udp.udp_ready = 0;
	udp.udp_size = cbAvail;
	udp.udp_fromip = pIpAddr->in_addr;
	udp.udp_fromport = pIpAddr->sin_port;
	/* hack--save location of UDP packet in circular buffer for callback */
	pCBM->cb_ready = pSock->s_recvbuf->cb_head;
	PutCBuf (pSock->s_recvbuf, (PCHAR) &udp, sizeof(udp));

	pCBM->cb_socket = pSock;
	if (pSock->s_recvbuf->cb_head + cbAvail >= pSock->s_recvbuf->cb_end) {
		/* wrap around */
		pCBM->cb_NB1.VirtualAddress = pSock->s_recvbuf->cb_head;
		pCBM->cb_NB1.Length = pSock->s_recvbuf->cb_end -
								pSock->s_recvbuf->cb_head;
		pCBM->cb_NB1.Next = &pCBM->cb_NB2;
		pCBM->cb_NB2.VirtualAddress = pSock->s_recvbuf->cb_start;
		pCBM->cb_NB2.Length = cbAvail - pCBM->cb_NB1.Length;
		pSock->s_recvbuf->cb_head = pSock->s_recvbuf->cb_start +
						pCBM->cb_NB2.Length;
	} else {
		/* no wrap */
		pCBM->cb_NB1.VirtualAddress = pSock->s_recvbuf->cb_head;
		pCBM->cb_NB1.Length = cbAvail;
		pSock->s_recvbuf->cb_head += cbAvail;
	}
	UnlockCBuf (pSock->s_recvbuf);

	*pcbTaken = 0;
	pCBM->cb_event.erb_buffer = &pCBM->cb_NB1;
	pCBM->cb_event.erb_size = cbAvail;
	pCBM->cb_event.erb_rtn = RcvCallback;
	pCBM->cb_event.erb_context = (PVOID) pCBM;
	pCBM->cb_event.erb_flags = &pCBM->cb_flags;

	*ppBuffer = &pCBM->cb_event;
	return TDI_MORE_PROCESSING;
}
